/*
 * Copyright (C) 2000-2014 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <portLib.h>
#include <fcntl.h>
#include <termios.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <ctype.h>

#include <taskLib.h>
#include <sysLib.h>

#include "h2mathLib.h"

#include "ptDriver.h"
#include "ptDriverPrivate.h"


/* ---------------------------- CONSTANTS  --------------------------- */


PT_CMD_DESC ptCmdDesc[] = PT_DEFAULT_CMD_DESCRIPTION;

const PT_ERR_FORMAT ptErrDesc[] = PT_ERR_DESC;

const PT_STRINGS_FORMAT ptStringsDesc[] = PT_STRINGS_DESC;


/* ------------------ LOCAL FUNCTION PROTOTYPES   -------------------- */

static int ptInitRS232(const char *device, int baudRate,
    PT_DRIVER_STR *ptDriverSDI);
static void ptDecode( PT_DRIVER_STR *ptDriverSDI);

static BOOL ptRecordReply(PT_CMD cmd, PT_CMD_STATUS replyStatus,
			  PT_ERR err, PT_DRIVER_STR *ptDriverSDI);
static BOOL ptRemoveCmdFromFifoList(PT_CMD cmd, PT_DRIVER_STR *ptDriverSDI);
static BOOL ptTestConst(PT_DRIVER_STR *ptDriverSDI);


/* ====================================================================== *
 *                          SYNCHRO                                     *
 * ====================================================================== */
#define PT_SEM_TAKE_TIMEOUT 1000 /* Tics */

BOOL
ptTakeSDI(const PT_DRIVER_STR *ptDriverSDI, char *name)
{
#ifdef VERBOSE
  printf ("%s TAKE...", name);
  fflush(stdout);
#endif
  if (semTake (ptDriverSDI->privateData->ptSDISem, PT_SEM_TAKE_TIMEOUT) != OK) {
    printf ("ptTakeSDI: %s cannot take sem\n", name);
    return FALSE;
  }
#ifdef VERBOSE
  printf ("TAKE OK\n");
#endif
  return TRUE;
}

void
ptGiveSDI(const PT_DRIVER_STR *ptDriverSDI, char *name)
{
#ifdef VERBOSE
  printf ("%s GIVE...", name);
  fflush(stdout);
#endif
  semGive(ptDriverSDI->privateData->ptSDISem);
#ifdef VERBOSE
  printf ("GIVE OK\n");
#endif
}

BOOL
ptUpdateVersion(PT_DRIVER_STR *ptDriverSDI, int v0, int v1, int v2)
{
  /* not very efficent but I don't know how to get the macro
     value in an other way .... */
  PT_CMD_DESC ptCmdDesc1_9[] = PT_CMD_DESCRIPTION_V_1_9;
  PT_CMD_DESC ptCmdDesc1_10[] = PT_CMD_DESCRIPTION_V_1_10;

  /* test major version number */
  if (v0 != 1) {
    fprintf (stderr, "%s: ptChangeVersion unkwnom version %d\n",
	     ptDriverSDI->ptDriverName, v0);
    return FALSE;
  }

  /* minor version changed */
  if (ptDriverSDI->v1 != v1) {

    /* version 1.9 */
    if (v1 == 9) {
      if (sizeof(ptCmdDesc) != sizeof(ptCmdDesc1_9)) {
	fprintf (stderr, "%s: ptChangeVersion failed\n",
		 ptDriverSDI->ptDriverName);
	return FALSE;
      }
      ptTakeSDI(ptDriverSDI, "ptDriver");
      memcpy(ptCmdDesc, ptCmdDesc1_9, sizeof(ptCmdDesc));
      ptGiveSDI(ptDriverSDI, "ptDriver");
    }

    /* version 1.10 */
    else if (v1 == 10) {
      ptDriverSDI->v1 = v1;
      if (sizeof(ptCmdDesc) != sizeof(ptCmdDesc1_10)) {
	fprintf (stderr, "%s: ptChangeVersion failed\n",
		 ptDriverSDI->ptDriverName);
	return FALSE;
      }
      ptTakeSDI(ptDriverSDI, "ptDriver");
      memcpy(ptCmdDesc, ptCmdDesc1_10, sizeof(ptCmdDesc));
      ptGiveSDI(ptDriverSDI, "ptDriver");
    }

    /* other */
    else {
      fprintf (stderr, "%s: ptChangeVersion unkwnom version %d.%d\n",
	       ptDriverSDI->ptDriverName, v0, v1);
      return FALSE;
    }

    printf ("%s: ptChangeVersion %d.%d -> %d.%d\n",
	    ptDriverSDI->ptDriverName,
	    ptDriverSDI->v0, ptDriverSDI->v1, v0, v1);
  }
  else {
    if (ptDriverSDI->ptVerbose) 
      printf ("%s: ptChangeVersion : version %d.%d unchanged\n",
	      ptDriverSDI->ptDriverName,
	      ptDriverSDI->v0, ptDriverSDI->v1);
  }

  /* record new version */
  ptDriverSDI->v0 = v0;
  ptDriverSDI->v1 = v1;
  ptDriverSDI->v2 = v2;

  return TRUE;
}

/* ====================================================================== *
 *                          THE DRIVER                                     *
 * ====================================================================== */

/*----------------------------------------------------------------------
 * ptDriver  -  server that reads PTU outputs
 *
 * One should uses ptInitDriver() to start this driver (see ptClient.c).
 *
 * This function waits for a string ended by '\n' or '\r' on the serial link.
 * The string is then stored in ptMsg and decoded using ptDecode().
 * If the message is identified, it is recorded in a list of the messages
 * using ptRecordReply(). This list will be consulted by the final client
 * of the PTU using adequat functions (see ptClient.c)
 */

void *
ptDriver(void *args)
{
  PT_DRIVER_ARGS *driverArgs = (PT_DRIVER_ARGS *)args;
  const char *device = driverArgs->device;
  int baudRate = driverArgs->baudRate;
  PT_DRIVER_STR *ptDriverSDI = driverArgs->sdiPtr;
  int count;
  int msgNum=0;
  char buf[PT_MAX_MSG_LENGTH];
  int status;
  char c;

  if (!ptTestConst(ptDriverSDI))
    return NULL;

  /* Initialization of the serial line */
  if ((ptDriverSDI->privateData->ptSerialFd = ptInitRS232(device, baudRate,
						      ptDriverSDI)) == -1) {
    fprintf (stderr, "%s: Error in init_RS232\n",
	     ptDriverSDI->ptDriverName);
    return NULL;
  }

  ptDriverSDI->ptStopDriverFlag = FALSE;

  /* MAIN READ LOOP */
  while (!ptDriverSDI->ptStopDriverFlag) {

     memset(buf, 0, PT_MAX_MSG_LENGTH);

    /* Reads message */
     count = 0;
     c = '\0';

    do {
      status = read(ptDriverSDI->privateData->ptSerialFd, &buf[count], 1);
      if (status == 1) {
	c  = buf[count];
	count++;
      }
      else {
	if (errno != 0 && errno != EINTR)
		fprintf(stderr, "%s: read(%d): %d %s\n",
			ptDriverSDI->ptDriverName,
			ptDriverSDI->privateData->ptSerialFd,
			status,
			strerror(errno));
	if (errno == 0) {

	  printf ("%s: no data. Is platine on ?\n",
		  ptDriverSDI->ptDriverName);
		  taskDelay(100);
	}
      }
    } while (c != '\n' && c != '\r' &&
	     !ptDriverSDI->ptStopDriverFlag && count<PT_MAX_MSG_LENGTH &&
	     /* Particular messages with no line return: !T et !P */
	     (count != 2 || (c != 'T' && c != 'P')));


    if (count == 1) {
      continue;
    }
    if (count == PT_MAX_MSG_LENGTH) {
      fprintf (stderr, "%s: Message too long\n", ptDriverSDI->ptDriverName);
    }

    /*==== TAKE SEMAPHORE =====*/
    if (!ptTakeSDI(ptDriverSDI, "ptDriver"))
	ptDriverSDI->ptStopDriverFlag = TRUE;

    /* Saves complet message */
    if (!ptDriverSDI->privateData->ptMsgFlag)
      ptDriverSDI->privateData->ptMsgFlag = TRUE;

    /* Removes \n or \r and add \0 */
    strncpy(ptDriverSDI->privateData->ptMsg, buf, count);
    if (buf[count-1] == '\n' || buf[count-1] == '\r')
      ptDriverSDI->privateData->ptMsg[count-1] = '\0';
    else
      ptDriverSDI->privateData->ptMsg[count] = '\0';

    if (ptDriverSDI->ptVerbose)
      printf ("%s: MSG %d: %s\n", ptDriverSDI->ptDriverName,
	      msgNum++, ptDriverSDI->privateData->ptMsg);

    /* Decodes message */
    ptDecode(ptDriverSDI);

    /*==== GIVE SEMAPHORE ====*/
    ptGiveSDI(ptDriverSDI, "ptDriver");

  } /* while */


  printf ("ptDriver stopped. Closing serial. Bye\n");
  close(ptDriverSDI->privateData->ptSerialFd);
  ptDriverSDI->ptDriverInitialized = FALSE;

  return NULL;
}

static BOOL
ptTestConst(PT_DRIVER_STR *ptDriverSDI)
{
  int ptNbCmd, ptNbErrMsg;
  int i;

  /* Coherency tests */
  ptNbCmd = sizeof(ptCmdDesc)/sizeof(PT_CMD_DESC);
  if (ptNbCmd != PT_NB_CMD)
    fprintf (stderr, "%s: WARNING : bug !: ptNbCmd %d != PT_NB_CMD %d\n",
	     ptDriverSDI->ptDriverName, ptNbCmd, PT_NB_CMD);

  ptNbErrMsg = sizeof(ptErrDesc)/sizeof(PT_ERR_FORMAT);
  if (ptNbErrMsg != PT_NB_ERR)
    fprintf (stderr, "%s: WARNING : bug !: ptNbErrMsg %d != PT_NB_ERR %d\n",
	    ptDriverSDI->ptDriverName, ptNbErrMsg, PT_NB_ERR);

  /* Tests command number attribution consistency
     (ie, ptCmdDesc[i].num == i) */
  for(i=0; i<PT_NB_CMD; i++) {
    if (ptCmdDesc[i].num != i) {
      fprintf (stderr,
	       "%s: INCONSISTENCY ERROR: the command number %d (named %s)"
	       " should be equal to its index %d in the array ptCmdDesc."
	       " Please check PT_CMD in ptDriver.h\n",
	       ptDriverSDI->ptDriverName, ptCmdDesc[i].num,
	       ptCmdDesc[i].name, i);
      return FALSE;
    }
  }
  /* Idem pour les erreurs */
  for(i=0; i<PT_NB_ERR; i++) {
    if (ptErrDesc[i].num != i) {
      fprintf (stderr,
	       "%s: INCONSISTENCY ERROR: the error number %d (named %s)"
	       " should be equal to its index %d in the array ptErrDesc[]."
	       " Please check PT_ERR in ptDriver.h\n",
	       ptDriverSDI->ptDriverName, ptErrDesc[i].num,
	       ptErrDesc[i].name, i);
      return FALSE;
    }
  }
  /* Idem pour les strings */
  for(i=0; i<PT_NB_STRINGS_DECODE; i++) {
    if (ptStringsDesc[i].num != i) {
      fprintf (stderr,
	       "%s: INCONSISTENCY ERROR: the string number %d (named %s)"
	       " should be equal to its index %d in the array ptStringsDesc[]."
	       " Please check PT_STRINGS_DECODE in ptDriver.h\n",
	       ptDriverSDI->ptDriverName,
	       ptStringsDesc[i].num,
	       ptStringsDesc[i].name, i);
      return FALSE;
    }
  }
  return TRUE;
}

/*-----------------------------------------------------------------------
 * ptDecode - decodes messages incoming from PTU.
 *
 * This function is called by ptDriver().
 *
 * Only messages that match a command sent using the function ptSendCmd()
 * (nbWaitedReplies != 0) will be recorded in the answers list.
 *
 * Once the answer of a command is identified, then nbWaitedReplies
 * is decremented and nbReceivedReplies is incremented.
 *
 * If a value is associated with the message, it is stored in ptDataVal[cmd].
 *
 * Unidentified messages are lost (there is only a printf).
 *
 * WARNING :
 *
 * 1. The PTU does not explicitely reference the command in its answer.
 *    Thus only the commands that return particular answer can be identified.
 *    In case of ambiguity, the answer is attributed to the first command
 *    with compatible answer format in FIFO order.
 *
 * 2. For similar reason error messages cannot be attributed to a
 *    particular command.
 *
 */

static void ptDecode(PT_DRIVER_STR *ptDriverSDI)
{
  int i, cmd;
  int index=0;
  BOOL found;
  PT_DATA_VAL val;
  char *p = &(ptDriverSDI->privateData->ptMsg[0]);

  /* Nettoyage */
  while (*p == ' ')
    p++;

  /**
   ** Stop index
   **/
  if (strncmp(p, "!T", 2) == 0) {
    if (ptDriverSDI->ptVerbose)
      printf ("%s: Tstop\n", ptDriverSDI->ptDriverName);
    ptDriverSDI->privateData->ptMsgFlag = FALSE;
    return;
  }
  else if (strncmp(p, "!P", 2) == 0) {
    if (ptDriverSDI->ptVerbose)
      printf ("%s: Pstop\n", ptDriverSDI->ptDriverName);
    ptDriverSDI->privateData->ptMsgFlag = FALSE;
    return;
  }

  /**
   ** Error message
   **/
  else if (strncmp(p, "! ", 2) == 0) {

    /* Print and record */
    strcpy(ptDriverSDI->privateData->ptErrMsg, p+2);

    /* Analyses */
    ptDriverSDI->privateData->ptErr = PT_UNKNOWN_ERR;
    for (i=0; i<PT_NB_ERR_FORMATS; i++) {
      if (strncmp(ptDriverSDI->privateData->ptErrMsg, ptErrDesc[i].format,
		  strlen(ptErrDesc[i].format)) == 0) {
	fprintf (stderr, "%s: ERR MSG : %s -> %s\n",
		 ptDriverSDI->ptDriverName,
		 ptDriverSDI->privateData->ptErrMsg, ptErrDesc[i].name);
	ptDriverSDI->privateData->ptErr = i;
	break;
      }
    }
    /* not identified */
    if (ptDriverSDI->privateData->ptErr == PT_UNKNOWN_ERR)
      fprintf (stderr, "%s: UNKNOWN ERR MSG : %s\n",
	       ptDriverSDI->ptDriverName,
	       ptDriverSDI->privateData->ptErrMsg);

    /* attributed to first command in fifo list */
    cmd = ptDriverSDI->privateData->ptWaitedRepliesFifoList[0];
    ptRecordReply(cmd, PT_REPLY_ERROR, ptDriverSDI->privateData->ptErr, ptDriverSDI);
    if (ptDriverSDI->privateData->ptErr == PT_UNKNOWN_ERR)
      strncpy(ptDriverSDI->privateData->ptDataVal[cmd].sval, ptDriverSDI->privateData->ptErrMsg, PT_MAX_STRING-1);

    ptDriverSDI->privateData->ptMsgFlag = FALSE;
    return;
  }
  else if (ptDriverSDI->privateData->ptMsg[0] == '*') {
    index = 1;
    while (ptDriverSDI->privateData->ptMsg[index] == ' ')
      index++;
  }

  /**
   ** Normal message
   **/

    /* Look round waiting command messages */
  for (i=0; i<ptDriverSDI->privateData->ptNbWaitedRepliesFifoList
	 && ptDriverSDI->privateData->ptMsgFlag; i++) {

    found = FALSE;
    cmd = ptDriverSDI->privateData->ptWaitedRepliesFifoList[i];

    /* Answer attribution + returned value record (if any) */
    switch (ptCmdDesc[cmd].outputDataType) {
    case PT_UNKNOWN_DATA:
      found = 1;
      printf ("%s: ptDecode answer cmd %s: %s\n",
	      ptDriverSDI->ptDriverName,
	      ptCmdDesc[cmd].name,
	      ptDriverSDI->privateData->ptMsg);
      break;

    case PT_INT_DATA:
      if ((found = sscanf(&ptDriverSDI->privateData->ptMsg[index],
			  ptCmdDesc[cmd].outputFormat,
			  &val.ival))==1)
	ptDriverSDI->privateData->ptDataVal[cmd].ival = val.ival;
      break;
    case PT_DOUBLE_DATA_S:
      if (sscanf(&ptDriverSDI->privateData->ptMsg[index], "%lf", &val.dval)==1) {
	while (ptDriverSDI->privateData->ptMsg[index] != ' ' &&
	       index<PT_MAX_MSG_LENGTH)
	  index ++;
	while (ptDriverSDI->privateData->ptMsg[index] == ' '&&
	       index<PT_MAX_MSG_LENGTH)
	  index ++;
	if (strncmp(&(ptDriverSDI->privateData->ptMsg[index]),
		    ptCmdDesc[cmd].outputFormat,
		    sizeof(ptCmdDesc[cmd].outputFormat)) == 0) {
	  ptDriverSDI->privateData->ptDataVal[cmd].dval = val.dval;
	  found = TRUE;
	}
      }
      break;
    case PT_STRING_DATA:
      if ((found = sscanf(&ptDriverSDI->privateData->ptMsg[index],
			  ptCmdDesc[cmd].outputFormat, val.sval))==1)
	strcpy(ptDriverSDI->privateData->ptDataVal[cmd].sval, val.sval);
      break;
    case PT_NO_DATA:
      found = (ptDriverSDI->privateData->ptMsg[index] == '\0');
      break;
    case PT_STRING_NO_DATA:
      found = (strncmp(&ptDriverSDI->privateData->ptMsg[index],
		       ptCmdDesc[cmd].outputFormat,
		       sizeof(ptCmdDesc[cmd].outputFormat)) == 0);
      break;
    case PT_NO_ANSWER:
      break;
    } /* switch */

    /* Command-answer association found */
    if (found == 1) {
      ptRecordReply(cmd, PT_REPLY_OK, PT_NO_ERR, ptDriverSDI);
      ptDriverSDI->privateData->ptMsgFlag = FALSE;
    }
  }

  /* Associated command not found */
  if (ptDriverSDI->privateData->ptMsgFlag && ptDriverSDI->ptVerbose) {
    printf ("%s: ptDecode: LONELY MSG %s\n", ptDriverSDI->ptDriverName,
	    ptDriverSDI->privateData->ptMsg);

    /* Spy on going cmds */
    ptSpyAll(ptDriverSDI);
    
    /* FIFO list of sent commands */
    printf ("%d answers waited (in chrono order): ",
	    ptDriverSDI->privateData->ptNbWaitedRepliesFifoList);
    for (i=0;i<ptDriverSDI->privateData->ptNbWaitedRepliesFifoList;i++) {
      printf (" %d ", ptDriverSDI->privateData->ptWaitedRepliesFifoList[i]);
    }
    printf ("\n");
  }
}


/*----------------------------------------------------------------------
 * ptRecordReply - Record a reply identified by ptDecode
 */

static BOOL ptRecordReply(PT_CMD cmd, PT_CMD_STATUS replyStatus,
			  PT_ERR err, PT_DRIVER_STR *ptDriverSDI)
{
  /* All commands ??? */
  if (cmd == PT_ALL /* -1 */) {
    ptDriverSDI->privateData->ptNbWaitedRepliesFifoList = 0;
  }

  /* Waited message ? */
  if (ptDriverSDI->privateData->ptCmdState[cmd].nbWaitedReplies<=0) {
    printf ("%s: ptRecordReply: no cmd %s %d on\n",
	    ptDriverSDI->ptDriverName, ptCmdDesc[cmd].name, cmd);
    return FALSE;
  }

  /* One less answer to be wait */
  ptDriverSDI->privateData->ptCmdState[cmd].nbWaitedReplies--;

  /* Not any more waited by the client */
  if (ptDriverSDI->privateData->ptCmdState[cmd].nbRepliesToRemove>0) {
    if (ptDriverSDI->ptVerbose)
      printf ("Removed reply : %d\n", cmd);
    ptDriverSDI->privateData->ptCmdState[cmd].nbRepliesToRemove--;
  }

  /* Still waited by the client */
  else
    ptDriverSDI->privateData->ptCmdState[cmd].nbReceivedReplies++;

  /* Answer status */
  ptDriverSDI->privateData->ptCmdState[cmd].replyStatus = replyStatus;
  ptDriverSDI->privateData->ptCmdState[cmd].replyErr = err;

  return ptRemoveCmdFromFifoList(cmd, ptDriverSDI);
}

/*----------------------------------------------------------------------
 * ptRemoveCmdFromFifoList -
 */

static BOOL ptRemoveCmdFromFifoList(PT_CMD cmd, PT_DRIVER_STR *ptDriverSDI)
{
  int i, j, k;

  /* Removes from FIFO list */
  for (i=0; i<ptDriverSDI->privateData->ptNbWaitedRepliesFifoList; i++) {

    /* Looks for corresponding command */
    if (ptDriverSDI->privateData->ptWaitedRepliesFifoList[i] == cmd) {

      /* Found : Trouvee: On l'ecrase remontant les suivantes d'une place */
      for(j=i; j<ptDriverSDI->privateData->ptNbWaitedRepliesFifoList-1; j++) {
	ptDriverSDI->privateData->ptWaitedRepliesFifoList[j]
	  = ptDriverSDI->privateData->ptWaitedRepliesFifoList[j+1];
      }

      /* Une de moins a attendre */
      (ptDriverSDI->privateData->ptNbWaitedRepliesFifoList)--;

      /* Tiens ? Ce n'etait pas la premiere de la liste FIFO ! */
      if (i != 0) {
	if (ptDriverSDI->ptVerbose) {
	  printf ("%s: %d not first. Waiting ",
		  ptDriverSDI->ptDriverName, cmd);
	  for (k=0;k<i;k++)
	    printf (" %d",
		    ptDriverSDI->privateData->ptWaitedRepliesFifoList[k]);
	  printf ("\n");
	}
      }

      /* Fin OK */
      return TRUE;
    }
  }

  /* Pas trouvee la commande dans la liste FIFO */
  printf ("%s: ptRemoveCmdFromFifoList: unknown cmd %d!\n",
	  ptDriverSDI->ptDriverName, cmd);
  return FALSE;
}


/*----------------------------------------------------------------------
 *  ptInitRS232  - Init de la ligne serie
 */
static struct speed {
    int speed;
    int termioSpeed;
} speeds[] = {
    { 0, B0 },
    { 50, B50 },
    { 75, B75 },
    { 110, B110 },
    { 300, B300 },
    { 600, B600 },
    { 1200, B1200 },
    { 2400, B2400 },
    { 4800, B4800 },
    { 9600, B9600 },
    { 19200, B19200 },
    { 38400, B38400 },
    { 57600, B57600 }};

#define SPEED_SIZE (sizeof(speeds)/sizeof(struct speed))

static int
get_termios_speed(int speed)
{
    int i;

    for (i = 0; i < SPEED_SIZE; i++) {
	if (speed == speeds[i].speed) {
	  return speeds[i].termioSpeed;
	}
    } /* for */
    return -1;
}

static int
ptInitRS232(const char *device, int baudRate, PT_DRIVER_STR *ptDriverSDI)
{
  int serialFd;
  struct termios tio;
  int speed;

  serialFd = open(device, O_RDWR);

  if (serialFd == -1) {
    fprintf(stderr, "ptInitRS232 error open %s", device);
    perror(":");
    return -1;
  }
  if (tcgetattr(serialFd, &tio) < 0) {
    fprintf(stderr, "ptInitRS232 while setting options ");
    perror("tcgetattr");
    return -1;
  }

  speed = get_termios_speed(baudRate);
  if (speed == -1) {
    fprintf(stderr, "ptInitRS232 error set baud rate %d\n", baudRate);
    return -1;
  }
  cfmakeraw(&tio);
  cfsetispeed(&tio, speed);
  cfsetospeed(&tio, speed);

  /* wait 0.5 sec for characters after the 1st one */
  tio.c_cc[VTIME] = 5;

  if (tcsetattr(serialFd, TCSAFLUSH, &tio) < 0) {
    fprintf(stderr, "ptInitRS232 while setting options ");
    perror("tcsetattr");
    return -1;
  }
  if (ptDriverSDI->ptVerbose) {
      printf ("\nptInitRS232: Connected to serial line %s, fd=%d\n",
	      device, serialFd);
  }
  ptDriverSDI->privateData->ptSerialFd = serialFd;

  ptDriverSDI->v0 = PT_DEFAULT_VERSION_V0;
  ptDriverSDI->v1 = PT_DEFAULT_VERSION_V1;
  ptDriverSDI->v2 = PT_DEFAULT_VERSION_V2;

  return (serialFd);
}
